package inputport.datacomm.simplex.buffer.nio;

import java.io.IOException;
import java.nio.ByteBuffer;
import java.nio.channels.SelectableChannel;
import java.nio.channels.SelectionKey;
import java.nio.channels.SocketChannel;
import java.util.HashMap;
import java.util.Map;
import java.util.concurrent.ArrayBlockingQueue;
import java.util.concurrent.BlockingQueue;

import common.Tracer;


public class AScatterGatherSelectionManager extends ASelectionManager implements MessagingSelectingRunnable  {
	public static final int BYTES_IN_INT =  Integer.SIZE/Byte.SIZE;
	BlockingQueue<BufferedHeaderWrite> headerBufferPool = new ArrayBlockingQueue<BufferedHeaderWrite>(MAX_BUFFERS);
	Map<SelectableChannel, Boolean> channelToExpectingHeader = new HashMap();
	Map<SelectableChannel, Boolean> channelToPartialMessageReceived = new HashMap();
	Map<SelectableChannel, Integer> channelToMessageLength = new HashMap();

	int messageLength;

	// assumes that the tables below are accessed after getReadBuffer() is called
	@Override
	protected void allocateReadState(SocketChannel theSocketChannel) {
		super.allocateReadState(theSocketChannel);
		channelToExpectingHeader.put(theSocketChannel, true);
		channelToPartialMessageReceived.put(theSocketChannel, false);
		channelToMessageLength.put(theSocketChannel, 0);		
	}
	public AScatterGatherSelectionManager() {
		allocateHeaderBuffers();
	}	
	void allocateHeaderBuffers() {
		Tracer.info(this, "Allocating resuable header read buffers for message sizes");
		for (int i = 0; i < MAX_BUFFERS; i++) {
			add(new ABufferedHeaderWrite(this, null, null, null));
		}
	}
	public void add (BufferedHeaderWrite theBufferedHeaderWrite) {
		try {
			headerBufferPool.put(theBufferedHeaderWrite);
		} catch (Exception e) {
			e.printStackTrace();
		}
	}
	@Override
	public synchronized void putBufferedWrite (BufferedWrite bufferedWrite) {
		try {
			// called by client, so can block
			BufferedHeaderWrite headerBufferedWrite = headerBufferPool.take();
			headerBufferedWrite.makeHeaderFor(bufferedWrite);
			BlockingQueue<BufferedWrite> bufferedWriteBoundedBuffer = getBufferedWriteBoundedBuffer(bufferedWrite);
			Tracer.info(this, "Qeuing with selector header buffer:" +  headerBufferedWrite);
			bufferedWriteBoundedBuffer.put(headerBufferedWrite);
			super.putBufferedWrite(bufferedWrite)	;		
		} catch (Exception e) {
			e.printStackTrace();
		}
	}
	
	@Override
	protected void processRead(SelectionKey selectionKey) throws IOException {
		SocketChannel socketChannel = (SocketChannel) selectionKey.channel();
		readIntoBuffer(socketChannel);
	    ByteBuffer readBuffer = getReadBuffer(socketChannel);
	    channelToPartialMessageReceived.put(socketChannel, false);// have received new data, so older value may not be correct

		while (true) {
			if (readBuffer.remaining() == 0 || channelToPartialMessageReceived.get(socketChannel)) {
				if (!maybeClearBuffer(readBuffer,channelToPartialMessageReceived.get(socketChannel), channelToExpectingHeader.get(socketChannel) ))
				resetToWritePastLimit(readBuffer);
				break;				
			} if (channelToExpectingHeader.get(socketChannel)) {
				extractMessageLength(socketChannel);
			} else {
				maybeExtractMessageAndNotifyReadListener(socketChannel);
			}
		}
		// reset if no state is required, the readBuffer check is probably redundant given the loop above
	
	}
	protected static void resetToReadFromMark(ByteBuffer aByteBuffer) {
		int oldPosition = aByteBuffer.position();
		if (oldPosition == 0) // no mark
			return;
		aByteBuffer.reset();
		aByteBuffer.limit(oldPosition);		
	}
	protected static void resetToWritePastLimit(ByteBuffer theByteBuffer) {
		theByteBuffer.position(theByteBuffer.limit());
		theByteBuffer.limit(theByteBuffer.capacity());
	}
	protected boolean maybeClearBuffer(ByteBuffer aByteBuffer, boolean partialMessageReceived, boolean expectingHeader) {
		boolean clearBuffer = aByteBuffer.remaining() == 0 &&
							  !partialMessageReceived && 
							  expectingHeader;
		if (clearBuffer) {
			Tracer.info(this, "Clearing read buffer as all bytes of message received and consumed");
			aByteBuffer.clear();
			aByteBuffer.mark();
		}
		return clearBuffer;
	}
	protected void extractMessageLength (SocketChannel socketChannel) {
		try {
		    ByteBuffer readBuffer = getReadBuffer(socketChannel);
			int totalBytesRead = readBuffer.remaining();
			while (totalBytesRead < BYTES_IN_INT) {
				Tracer.info(this, "Header bytes read less than header size, polling for remaining bytes");
				resetToWritePastLimit(readBuffer);
				Tracer.info(this, "Set position to limit, and limit to capacity:" + readBuffer);				
				int nextBytesRead = readIntoBuffer(socketChannel);

				totalBytesRead += nextBytesRead;
			
			}
			channelToExpectingHeader.put(socketChannel, false);
			messageLength = readBuffer.getInt();
			Tracer.info(this, "Message length extracted:" + messageLength);
			readBuffer.mark();
			Tracer.info(this, "Marked buffer at current position:" + readBuffer);
		} catch (Exception e) {
			e.printStackTrace();
		}		
	}	

	protected void maybeExtractMessageAndNotifyReadListener(SocketChannel socketChannel) {
	    ByteBuffer readBuffer = getReadBuffer(socketChannel);
		Tracer.info(this, "Read message buffer:" + readBuffer + " from " + socketChannel);
		if (readBuffer.remaining() < messageLength) {
			Tracer.info(this, "Did not receive full message length:" + messageLength);	
			Tracer.info(this, "Not changing position or limit");
			channelToPartialMessageReceived.put(socketChannel, true);
			return;
		}
		Tracer.info(this, "Received full message");
		channelToExpectingHeader.put(socketChannel, true);
		channelToPartialMessageReceived.put(socketChannel, false);		
		ByteBuffer messageBuffer = ByteBuffer.wrap(readBuffer.array(), readBuffer.position(), messageLength);
		Tracer.info(this, "Sendling clients new message byte buffer  sharing the array of the read buffer:" + messageBuffer);		
		notifyRead(socketChannel, messageBuffer, messageLength);
		// should we not clear the buffer?
		readBuffer.position(readBuffer.position() + messageLength);
		readBuffer.mark();
		Tracer.info(this, "Advanced read buffer position to first unconsumed byte and marked position:" + readBuffer);		


	}
	@Override
	protected int  readIntoBuffer(SocketChannel socketChannel) throws IOException {
		int bytesRead = super.readIntoBuffer(socketChannel);
	    ByteBuffer readBuffer = getReadBuffer(socketChannel);
		resetToReadFromMark(readBuffer);
		Tracer.info(this, "Resetting buffer  to mark and limit to position:" + readBuffer);
		return bytesRead;		
	}
}
